/*
** BeastAttack
** Author: Jiannan Zhang
* 
** The returned length of ciphertext is 56, includes 8 bytes IV, leaves 48 bytes, 6 blocks information.
** IV prediction: IV has a linear relation to system time, however the system time can only be obtained from outside the funciton, there is exact prediction to IV, so must guess. The approach I used is to call the encrypt program twice, and calculate the time difference between the time before first call and the time before second call. On the Linux systems in Oxford CS labs, the difference is either 1, 2 or 3 ms, and by observation, a certain number (~+4 for 1 ms, ~+8 for 2 ms, ~+15 for 3 ms), the IV for second call can be estimated, then after the second call, it can be verified. This observation works well but with variations at different time and on different computers, but on average 30% -- 50% hits (predicated IV is equal to the actual IV) can be achieved.
*
** From task 2, the first 8 bytes of the plaintext are: "The name" in ASCII. The performance could be improved by the fact that all characters are english characters, numbers and normal characters such as comma, peroid and so on, but this implementation searches for the whole space. Change is easy to do in this case.
*
** To decrypt the whole message, at any block i which has been decrypted, block i+1 can be decrypted from block i. The algorithm takes two runs, first to get a reference ci and ci+1, then it runs again with 8-byte prefix, put c1 into the prefix, and cancel IV, mi out, then guess the last byte. Repeat the procedure gets the entire plaintext message.
*/

import java.io.*;
import javax.xml.bind.annotation.adapters.HexBinaryAdapter;
import java.util.Arrays;
import java.nio.*;
import java.lang.*;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;

public class BeastAttack
{
    public static long beginTimeFirstRun=0;
    public static long beginTimeSecondRun=0;

    public static void main(String[] args) throws Exception
    {
	byte[] message=new byte[1024];				//To store message
	byte[] ciphertextSecondRun=new byte[1024];		//Cipher text of second run
	byte[] ciphertextFirstRun=new byte[1024];		//Cipher text of first run
	//First run, to get rid of irregular time intervals
	int length=callEncrypt(null, 0, ciphertextFirstRun);

	//Original code for displaying length and content of ciphertext
/*	System.out.println(String.format("Length of cipher is: %d\n",length));
	for(int i=0;i<length;i++)
	{
		System.out.print(String.format("%02x ",ciphertextFirstRun[i]));
	}*/
	
	//Keep 16 bytes of ciphertext when IV has a match
	byte[] referenceCiphertext=new byte[8];
	byte predictionIV8=0;
	byte[] IVs=new byte[8];
	byte[] referenceOldCiphertext=new byte[8];
	byte oldcipher8=0;
	 
	//For statistics
	int misstime=0;
	int equaltime=0;
	
	//flag to indicate the stage of current decryption
	int guess=0;		//1: to guess the byte
	int byteguess=0;	//location of current guess 0 -- 255
	int bytelocation=0;	//how many bytes have been moved in 0 -- 8
	int message_location=0;	//current location of decrypted message
	int showflag=0;		//for display message on the fly
	int blocklocation=1;	//position of block
	
	//Test things
	long oldIV=0;
	
	//Loop through the blocks
	while(blocklocation<7)
	{
		// e.g. this prints out the result of an encryption with no prefix	    
		while(bytelocation<=7)
		{
			byte prefix []=new byte[8];						//prefix to each cipher
			byte temparraycheck []=new byte[8];					//when guessing, use this arrary to compare
			//Prediction to IV for each run
			long predictionIV=0;
			ByteBuffer predictionIVBuf=ByteBuffer.wrap(prefix); 
			if(showflag==0)
			{
				System.out.print(String.format("%d Bytes Have Been Decrypted -----------",message_location));
				if(message_location>0)
				{
					System.out.println(String.format(" %c", message[message_location-1]));
				}
				else
				{
					System.out.println();
				}
				showflag=1;
			}
			//first run to locate system time
			beginTimeFirstRun=System.currentTimeMillis();
			length=callEncrypt(null, 0, ciphertextFirstRun);
			beginTimeSecondRun=System.currentTimeMillis();
			long difference=beginTimeSecondRun-beginTimeFirstRun;
		
			ByteBuffer bufIVs=ByteBuffer.wrap(IVs);
			
			//Predict IV
			if(difference>3 || difference<1)
			{
				continue;
			}
			else
			{
				ByteBuffer buffer=ByteBuffer.wrap(ciphertextFirstRun,0,8);
				long tempIV=buffer.getLong();
				if(difference==3)
		        	{
		                	predictionIV=tempIV+13;
		//			System.out.println("Case: 3");
		        	}
		        	else if(difference==2)
		        	{
					predictionIV=tempIV+7;
	     //                   	System.out.println("Case: 2");
		        	}
		        	else if(difference==1)
		        	{
					predictionIV=tempIV+3;
		//                	System.out.println("Case: 1");
		       		}

				if(guess==0)	//7-byte prefix with m1 xor IV8
				{
					if(blocklocation==1)
					{
						predictionIVBuf.putLong(0,predictionIV).array();
		                        	length=callEncrypt(prefix, 7-bytelocation, ciphertextSecondRun);
					}
					else
					{
						predictionIVBuf.putLong(0,predictionIV).array();
						length=callEncrypt(prefix, 7-bytelocation, ciphertextSecondRun);
					}
				}
				else
				{
					if(byteguess<256)
					{
						if(blocklocation==1)
						{
							predictionIVBuf.putLong(0,predictionIV).array();
							prefix[7]=(byte)(byteguess & 0xff);
							for(int i=7-bytelocation;i<7;i++)
							{
								prefix[i]^=(byte)(message[i+bytelocation-7])^IVs[i];
							}
							length=callEncrypt(prefix, 8, ciphertextSecondRun);
						}
						else
						{
							predictionIVBuf.putLong(0,predictionIV).array();
							for(int i=0;i<7;i++)
							{
								prefix[i]^=(byte)(message[(blocklocation-2)*8+bytelocation+1+i])^(referenceOldCiphertext[i]&0xff);;
							}
							prefix[7]^=(byte)(byteguess & 0xff)^oldcipher8;
							length=callEncrypt(prefix, 8, ciphertextSecondRun);
						}
					}
					else
					{
						guess=0;
						byteguess=0;
						continue;
					}
				}
	
				ByteBuffer bufVerify=ByteBuffer.wrap(ciphertextSecondRun,0,8);
				long verify=bufVerify.getLong();
				//IV predicted correctly, proceed
				if(verify==predictionIV)
				{
					equaltime++;
					//start trying to decrypt
					if(guess==0)
					{
						if(blocklocation==1)
						{
							guess=1;
							predictionIV8=(byte)(predictionIV&0xff);
							bufIVs.putLong(predictionIV);
							System.arraycopy(ciphertextSecondRun,8,referenceCiphertext,0,8);
							oldIV=predictionIV;
						}
						else
						{
							guess=1;
							System.arraycopy(ciphertextSecondRun,8*(blocklocation-1),referenceOldCiphertext,0,8);
							System.arraycopy(ciphertextSecondRun,8*blocklocation,referenceCiphertext,0,8);
							oldcipher8=(byte)(referenceOldCiphertext[7]&0xff);
							predictionIV8=(byte)(predictionIV&0xff);
							bufIVs.putLong(predictionIV);
							oldIV=predictionIV;
						}
					}
					else
					{
						System.arraycopy(ciphertextSecondRun,8,temparraycheck,0,8);
						if(Arrays.equals(referenceCiphertext,temparraycheck))
						{
							if(blocklocation==1)
							{
								int test=(predictionIV8&0xff)^(byteguess&0xff)^((int)predictionIV&0xff);
								message[message_location]=(byte)(test&0xff);
								message_location++;
								bytelocation++;
								guess=0;
								byteguess=0;
								showflag=0;
							}
							else
							{
								int test=byteguess&0xff;
								message[message_location]=(byte)(test&0xff);
								message_location++;
								bytelocation++;
								guess=0;
								byteguess=0;
								showflag=0;
							}
						}
						else if(byteguess<256)
						{
							byteguess++;
						}
						else
						{
							break;
						}
					}
				}	
				else
				{
					misstime++;
				}
			}
		}
		blocklocation++;
		bytelocation=0;
	}

	System.out.println(String.format("Hit: %d, Missed: %d\n",equaltime, misstime));	  
	System.out.println("The message is:");
	for(int i=0;i<48;i++)
	{
		System.out.print(String.format("%c",message[i]));
	}
	System.out.println();
    }
    


    // a helper method to call the external programme "encrypt" in the current directory
    // the parameters are the plaintext, length of plaintext, and ciphertext; returns length of ciphertext
    static int callEncrypt(byte[] prefix, int prefix_len, byte[] ciphertext) throws IOException
    {
	HexBinaryAdapter adapter = new HexBinaryAdapter();
	Process process;
	
	// run the external process (don't bother to catch exceptions)
	if(prefix != null)
	{
	    // turn prefix byte array into hex string
	    byte[] p=Arrays.copyOfRange(prefix, 0, prefix_len);
	    String PString=adapter.marshal(p);
	    process = Runtime.getRuntime().exec("./encrypt "+PString);
	}
	else
	{
	    process = Runtime.getRuntime().exec("./encrypt");
	}

	// process the resulting hex string
	String CString = (new BufferedReader(new InputStreamReader(process.getInputStream()))).readLine();
	byte[] c=adapter.unmarshal(CString);
	System.arraycopy(c, 0, ciphertext, 0, c.length); 
	return(c.length);
    }
}
